import re
import json

import pytest


def assert_get(url, expected_path, expected_params):
    import urllib.parse

    parsed_url = urllib.parse.urlparse(url)

    assert parsed_url.path == expected_path
    assert urllib.parse.parse_qs(parsed_url.query) == expected_params


@pytest.fixture(autouse=True)
def mock_all_triage_requests(requests_mock, fixture_from_file):
    requests_mock.get(
        re.compile("https://triage.example.com/"),
        text=fixture_from_file("empty.json"),
    )
    requests_mock.post(
        re.compile("https://triage.example.com/"),
        text=fixture_from_file("reports.json"),
    )


def test_get_reporters_by_email(requests_mock, triage):
    list(triage.get_reporters_by_email("reporter1@example.com"))

    assert_get(
        requests_mock.last_request.url,
        "/api/public/v2/reporters",
        {"filter[email]": ["reporter1@example.com"]},
    )


def test_get_resources_no_filter(mocker, triage):
    mocker.patch("cofense_triage.triage_api_client.TriageApiClient.get_documents")

    triage.get_reporters()

    triage.api_client.get_documents.assert_called_with("reporters", [])


def test_get_resources_filter(mocker, triage):
    mocker.patch("cofense_triage.triage_api_client.TriageApiClient.get_documents")

    triage.get_reporters([{"attr": "name", "val": "ggg"}])

    triage.api_client.get_documents.assert_called_with(
        "reporters", [{"attr": "name", "val": "ggg"}]
    )


def test_nonexistant_function(triage):
    with pytest.raises(AttributeError):
        triage.nonexistent_resource_type()


def test_create_resources_list_of_attrs(mocker, triage):
    mocker.patch("cofense_triage.triage_api_client.TriageApiClient.create_documents")

    triage.create_threat_indicators([{"a": 5, "b": 6}, {"a": 7, "b": 8}])

    triage.api_client.create_documents.assert_called_with(
        "threat_indicators", [{"a": 5, "b": 6}, {"a": 7, "b": 8}]
    )


def test_create_resources_kwargs(mocker, triage):
    mocker.patch("cofense_triage.triage_api_client.TriageApiClient.create_documents")

    triage.create_threat_indicators(a=5, b=6)

    triage.api_client.create_documents.assert_called_with(
        "threat_indicators", [{"a": 5, "b": 6}]
    )


def test_create_resources_list_and_kwargs(mocker, triage):
    mocker.patch("cofense_triage.triage_api_client.TriageApiClient.create_documents")

    triage.create_threat_indicators([{"a": 5, "b": 6}], a=7, b=8)

    triage.api_client.create_documents.assert_called_with(
        "threat_indicators", [{"a": 5, "b": 6}, {"a": 7, "b": 8}]
    )


def test_create_threat_indicator(requests_mock, triage):
    triage.create_threat_indicators(
        threat_level="Suspicious", threat_type="Sender", threat_value="evil@example.com"
    )

    # assert requests_mock.call_count == 1
    assert (
        requests_mock.last_request.url
        == "https://triage.example.com/api/public/v2/threat_indicators"
    )
    assert json.loads(requests_mock.last_request.body) == {
        "data": {
            "type": "threat_indicators",
            "attributes": {
                "threat_level": "Suspicious",
                "threat_type": "Sender",
                "threat_value": "evil@example.com",
            },
            "relationships": {},
        }
    }



import json


def test_parse(requests_mock, triage, fixture_from_file):
    requests_mock.get(
        "https://triage.example.com/api/public/v2/threat_indicators",
        text=fixture_from_file("threat_indicators.json"),
    )

    threat_indicator = next(triage.get_threat_indicators())

    assert threat_indicator.resource_id == "1"
    assert threat_indicator.threat_type == "Sender"
    assert threat_indicator.threat_value == "zoe.watts@example.org"
    assert threat_indicator.created_at == "2020-09-22T05:52:21.907Z"

    assert (
        json.loads(threat_indicator.to_json())["attributes"]["threat_level"]
        == "Malicious"
    )



from cofense_triage.triage_api_client import TriageApiClient
from cofense_triage.errors import ReporterNotFoundError

from cofense_triage.models import build_resource_class


class Triage:
    def __init__(self, *, host, api_version, client_id, client_secret):
        self.api_client = TriageApiClient(
            host=host,
            api_version=api_version,
            client_id=client_id,
            client_secret=client_secret,
        )

    def __getattr__(self, name):
        if name.startswith("get_"):
            # TODO Python 3.9: name = name.removeprefix("create_")
            resource_name = name[4:]
            return self._get_resources_function(resource_name)
        elif name.startswith("create_"):
            # TODO Python 3.9: name = name.removeprefix("create_")
            resource_name = name[7:]
            return self._create_resources_function(resource_name)
        else:
            raise AttributeError(name)

    def _create_resources_function(self, resource_name):
        def _function(resources=[], **kwargs):
            if isinstance(resources, dict):
                resources = [resources]
            elif resources is None:
                resources = []

            if kwargs:
                resources.append(kwargs)

            return self.api_client.create_documents(resource_name, resources)

        return _function

    def _get_resources_function(self, resource_name):
        def _function(filter_params=[]):
            if not isinstance(filter_params, list):
                filter_params = [filter_params]

            return (
                build_resource_class(
                    resource_name, document, oauth_session=self.api_client.oauth_session
                )
                for document in self.api_client.get_documents(
                    resource_name, filter_params
                )
            )

        return _function

    def get_processed_reports(self, filter_params=[]):
        return self.get_reports(
            filter_params + [{"attr": "location", "val": "Processed"}]
        )

    def get_processed_reports_since(self, date, filter_params=[]):
        return self.get_processed_reports(
            filter_params + [{"attr": "created_at", "val": date, "op": "gt"}]
        )

    def get_processed_reports_by_reporter(self, address, filter_params=[]):
        reporters = list(self.get_reporters_by_address(address))

        if not len(reporters) == 1:
            raise ReporterNotFoundError(address)

        return self.get_reports(
            filter_params + [{"attr": "location", "val": "Processed"}],
            resource_type=f"reporters/{reporters[0].resource_id}/reports",
        )

    def get_reporters_by_email(self, address, filter_params=[]):
        if isinstance(address, list):
            address = ",".join(address)
        return self.get_reporters(filter_params + [{"attr": "email", "val": address}])

    def get_operators_by_email(self, address, filter_params=[]):
        if isinstance(address, list):
            address = ",".join(address)
        return self.get_operators(filter_params + [{"attr": "email", "val": address}])



import importlib
import inspect
import pkgutil

RESOURCE_CLASSES = {}


def build_resource_class(name, document, oauth_session=None):
    return RESOURCE_CLASSES[name](document, oauth_session)


for module_info in pkgutil.iter_modules(__path__, prefix=f"{__package__}."):
    module = importlib.import_module(module_info.name)

    if not hasattr(module, "RESOURCE_NAME"):
        continue

    defined_classes = inspect.getmembers(
        module,
        lambda member: inspect.isclass(member) and member.__module__ == module.__name__,
    )

    # There should only be one (name, class) tuple in the model module. Take the class.
    resource_class = defined_classes[0][1]

    RESOURCE_CLASSES[module.RESOURCE_NAME] = resource_class



import abc
import json
from jsonapi_client.relationships import SingleRelationship

from cofense_triage.models import build_resource_class


class BaseModel(abc.ABC):
    def __init__(self, document, oauth_session=None):
        self._oauth_session = oauth_session
        self.document = document

    def __getattr__(self, name):
        if name in self.document._relationships.keys_python():
            if isinstance(self.document.relationships[name], SingleRelationship):
                resource_type_name = self.document[name].type
                return build_resource_class(
                    resource_type_name,
                    self.document[name],
                    oauth_session=self._oauth_session,
                )
            else:
                # it's a to-many relationship, so return a generator
                return (
                    build_resource_class(
                        name,
                        resource,
                        oauth_session=self._oauth_session,
                    )
                    for resource in self.document[name]
                )
        else:
            return self.document[name]

    def __setattr__(self, name, value):
        if name in ["document", "_oauth_session"]:  # TODO name.startswith("_")
            return super().__setattr__(name, value)
        self.document[name] = value

    def delete(self):
        self.document.delete()

    def commit(self):
        self.document.commit()

    @property
    def resource_id(self):
        return self.document.id

    @property
    def attributes(self):
        return self.to_dict()["attributes"]

    def to_dict(self):
        return self.document.json  # This library-provided attr is actually a dict

    def to_json(self):
        return json.dumps(self.to_dict())



from cofense_triage.triage_api_client import TriageApiClient


class TestTriageApiClient:
    def test__build_auth_object(self, mocker, requests_mock):
        requests_mock.post(
            "https://some-triage-host/oauth/token",
            text="""
                {
                    "access_token": "great-access-token",
                    "token_type": "bearer",
                    "expires_in": 3600,
                    "refresh_token": "great-refresh-token"
                }
            """,
        )
        mock_auth = mocker.patch("cofense_triage.triage_api_client.OAuth2")

        TriageApiClient(
            host="https://some-triage-host",
            api_version=2,
            client_id="great_client_id",
            client_secret="great_client_secret",
        )

        mock_auth_args = mock_auth.call_args[1]
        assert mock_auth_args["client_id"] == "great_client_id"
        assert mock_auth_args["token"]["access_token"] == "great-access-token"

    def test__build_jsonapi_session(self, mocker):
        mock_jsonapi_session = mocker.patch("jsonapi_client.Session")
        mocker.patch(
            "cofense_triage.triage_api_client.TriageApiClient._build_auth_object",
            return_value=("mock-oauth-session-obj", "mock-oauth-auth-obj"),
        )

        TriageApiClient(
            host="https://some-triage-host",
            api_version=2,
            client_id="great_client_id",
            client_secret="great_client_secret",
        )

        mock_jsonapi_session.assert_called_once_with(
            "https://some-triage-host/api/public/v2",
            request_kwargs={"auth": "mock-oauth-auth-obj"},
            schema=mocker.ANY,
            use_relationship_iterator=True,
        )



import json


def test_parse(requests_mock, triage, fixture_from_file):
    requests_mock.get(
        "https://triage.example.com/api/public/v2/reports",
        text=fixture_from_file("reports.json"),
    )

    report = next(triage.get_reports())

    assert report.resource_id == "1"
    assert report.from_address == "zack.bins@example.net"
    assert report.match_priority == 3
    assert report.reported_at == "2020-09-22T05:52:20.859Z"
    assert (
        report.jpg_url
        == "https://triage.example.com/api/public/v2/reports/1/download.jpg"
    )

    assert (
        json.loads(report.to_json())["attributes"]["subject"]
        == "Eveniet voluptatem dolor rerum"
    )


def test_attachments(requests_mock, triage, fixture_from_file):
    requests_mock.get(
        "https://triage.example.com/api/public/v2/reports",
        text=fixture_from_file("reports.json"),
    )
    requests_mock.get(
        "https://triage.example.com/api/public/v2/reports/1/attachments",
        text=fixture_from_file("attachments.json"),
    )
    requests_mock.get(
        "https://triage.example.com/api/public/v2/attachment_payloads/1",
        text=fixture_from_file("attachment_payloads.json"),
    )

    report = next(triage.get_reports())

    attachment = next(report.attachments)
    assert attachment.filename == "taco_menu.pdf"

    attachment_payload = attachment.attachment_payload
    assert attachment_payload.mime_type == "application/pdf; charset=binary"



from cofense_triage.models.base_model import BaseModel


RESOURCE_NAME = "reports"


class Report(BaseModel):
    """
    Class representing a Triage report by an end-user of a suspicious message
    """

    @property
    def jpg_url(self):
        return self.document.links.self.url + "/download.jpg"

    @property
    def jpg(self):
        return self._oauth_session.get(self.jpg_url)

    def categorize(self, category_id, response_id, categorization_tags):
        pass

    @property
    def png_url(self):
        return self.document.links.self.url + "/download.png"

    @property
    def png(self):
        return self._oauth_session.get(self.png_url)



import jsonapi_client


class FilterParams(jsonapi_client.Filter):
    def __init__(self, params):
        """
        jsonapi_client.Filter takes conditions as kwargs, which only allows for attributes and
        values. We also want to support operations, so we'll replace the kwargs with a list of dicts
        having three keys: attr, val, op.

        examples:
            Filter({"attr": "name", "val": "Example"})
            Filter({"attr": "name", "val": "Ex", "op": "start"})
            Filter({"attr": "created_at", "val": "2000-01-01", "op": "gt"})
        """
        self._params = params

    def appended_query(self):
        return self.format_filter_query(self._params)

    def format_filter_query(self, params):
        def build_key(param):
            if "op" in param:
                return f"{param['attr']}_{param['op']}"

            return param["attr"]

        return "&".join(
            f"filter[{build_key(param)}]={param['val']}" for param in params
        )



from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session, OAuth2
import jsonapi_client

from cofense_triage.filter_params import FilterParams
from cofense_triage import TRIAGE_SCHEMA


class TriageApiClient:
    def __init__(self, *, host, api_version, client_id, client_secret):
        if not api_version == 2:
            raise "unsupported API version"

        self.host = host
        self.api_version = api_version
        self.client_id = client_id
        self.client_secret = client_secret

        self.oauth_session, self.oauth_auth = self._build_auth_object()
        self.jsonapi_session = self._build_jsonapi_session()

    def _build_auth_object(self):
        client = BackendApplicationClient(client_id=self.client_id)
        oauth_session = OAuth2Session(client=client)
        token = oauth_session.fetch_token(
            token_url=f"{self.host}/oauth/token",
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

        return (
            oauth_session,
            OAuth2(client_id=self.client_id, client=client, token=token),
        )

    def _build_jsonapi_session(self):
        # TODO we have to deal with token expiration ourselves. jsonapi_session should be able to take an OAuth2Session.
        return jsonapi_client.Session(
            f"{self.host}/api/public/v{self.api_version}",
            request_kwargs={"auth": self.oauth_auth},
            schema=TRIAGE_SCHEMA,
            use_relationship_iterator=True,
        )

    def get_documents(self, resource_type, filter_params=None):
        # TODO sort param
        if filter_params:
            return self.jsonapi_session.iterate(
                resource_type, FilterParams(filter_params)
            )

        return self.jsonapi_session.iterate(resource_type)

    def create_documents(self, resource_type, resources):
        for resource in resources:
            # Triage properties contain underscores, so use `fields` instead of expanding kwargs
            new_resource = self.jsonapi_session.create(resource_type, fields=resource)

            # The resource must be manually added to the session. Perhaps a bug in jsonapi_client.
            self.jsonapi_session.add_resources(new_resource)

        return self.jsonapi_session.commit()



import pytest

from cofense_triage import Triage


@pytest.fixture
def fixture_from_file():
    def read_fixture_from_file(fname):
        with open(f"test/fixtures/{fname}", "r") as file:
            return file.read()

    return read_fixture_from_file


@pytest.fixture
def triage(mock_oauth_token):
    return Triage(
        host="https://triage.example.com",
        api_version=2,
        client_id="some-client-id",
        client_secret="some-client-secret",
    )


@pytest.fixture
def mock_oauth_token(requests_mock):
    requests_mock.post(
        "https://triage.example.com/oauth/token",
        text="""
            {
                "access_token": "great-access-token",
                "token_type": "bearer",
                "expires_in": 3600,
                "refresh_token": "great-refresh-token"
            }
        """,
    )



import setuptools
import pathlib


def read_readme():
    with open(pathlib.Path(__file__).parent / "README.md") as f:
        return f.read()


setuptools.setup(
    name="cofense_triage",
    version="0.1",
    author="Eddie Lebow",
    author_email="oss@cofense.com",
    maintainer="Cofense, Inc.",
    maintainer_email="oss@cofense.com",
    url="https://github.com/cofense/triage-sdk-python",
    description="Python interface to the Cofense Triage API V2",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    packages=setuptools.find_packages(),
    package_data={
        "cofense_triage": ["schema.json"]
    },
    install_requires=[
        "requests_oauthlib",
        "jsonapi_client",
    ],
)



import json


def test_parse(requests_mock, triage, fixture_from_file):
    requests_mock.get(
        "https://triage.example.com/api/public/v2/reporters",
        text=fixture_from_file("reporters.json"),
    )

    reporter = next(triage.get_reporters())

    assert reporter.resource_id == "1"
    assert reporter.email == "darren@example.com"
    assert reporter.reputation_score == 15
    assert reporter.created_at == "2020-09-22T05:52:20.669Z"

    assert json.loads(reporter.to_json())["attributes"]["email"] == "darren@example.com"
